/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: CommonsModelerExtService.java 4389 2008-12-15 13:48:57Z alitokmen $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.jmx.impl;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.ow2.util.jmx.api.IBaseModelMBeanExt;
import org.ow2.util.jmx.api.ICommonsModelerExtService;
import org.ow2.util.jmx.api.IMBeanAttribute;
import org.ow2.util.jmx.api.IMBeanOperation;

/**
 *
 * @author missonng
 */
public class CommonsModelerExtService implements ICommonsModelerExtService {
    /**
     * The event service state.
     */
    private boolean available = false;

    /**
     * Map ids to MBeans.
     */
    private Map<String, IBaseModelMBeanExt> ids2mbeans = new HashMap<String, IBaseModelMBeanExt>();

    /**
     * Map attributes to regexps.
     */
    private Map<IMBeanAttribute, String> attributes2regexps = new HashMap<IMBeanAttribute, String>();

    /**
     * Map operations to regexps.
     */
    private Map<IMBeanOperation, String> operations2regexps = new HashMap<IMBeanOperation, String>();

    /**
     * Map MBeans to ids to attibutes.
     */
    private Map<IBaseModelMBeanExt, Map<String, Set<IMBeanAttribute>>> mbeans2ids2attributes =
        new HashMap<IBaseModelMBeanExt, Map<String, Set<IMBeanAttribute>>>();

    /**
     * Map MBeans to ids to operations.
     */
    private Map<IBaseModelMBeanExt, Map<String, Set<IMBeanOperation>>> mbeans2ids2operations =
        new HashMap<IBaseModelMBeanExt, Map<String, Set<IMBeanOperation>>>();

    /**
     * Map MBeans to attributes to ids.
     */
    private Map<IBaseModelMBeanExt, Map<IMBeanAttribute, Set<String>>> mbeans2attributes2ids =
        new HashMap<IBaseModelMBeanExt, Map<IMBeanAttribute, Set<String>>>();

    /**
     * Map MBeans to operations to ids.
     */
    private Map<IBaseModelMBeanExt, Map<IMBeanOperation, Set<String>>> mbeans2operations2ids =
        new HashMap<IBaseModelMBeanExt, Map<IMBeanOperation, Set<String>>>();

    /**
     * Start the service.
     */
    public synchronized void start() {
        if (!this.available) {
            this.available = true;

            for (Entry<IBaseModelMBeanExt, Map<IMBeanAttribute, Set<String>>> entry : this.mbeans2attributes2ids.entrySet()) {
                final IBaseModelMBeanExt mbean = entry.getKey();
                final Set<IMBeanAttribute> attributes = entry.getValue().keySet();

                for (IMBeanAttribute attribute : attributes) {
                    mbean.addAttribute(attribute);
                }
            }

            for (Entry<IBaseModelMBeanExt, Map<IMBeanOperation, Set<String>>> entry : this.mbeans2operations2ids.entrySet()) {
                final IBaseModelMBeanExt mbean = entry.getKey();
                final Set<IMBeanOperation> operations = entry.getValue().keySet();

                for (IMBeanOperation operation : operations) {
                    mbean.addOperation(operation);
                }
            }
        }
    }

    /**
     * Stop the service.
     */
    public synchronized void stop() {
        if (!this.available) {
            this.available = true;

            for (Entry<IBaseModelMBeanExt, Map<IMBeanAttribute, Set<String>>> entry : this.mbeans2attributes2ids.entrySet()) {
                final IBaseModelMBeanExt mbean = entry.getKey();
                final Set<IMBeanAttribute> attributes = entry.getValue().keySet();

                for (IMBeanAttribute attribute : attributes) {
                    mbean.removeAttribute(attribute);
                }
            }

            for (Entry<IBaseModelMBeanExt, Map<IMBeanOperation, Set<String>>> entry : this.mbeans2operations2ids.entrySet()) {
                final IBaseModelMBeanExt mbean = entry.getKey();
                final Set<IMBeanOperation> operations = entry.getValue().keySet();

                for (IMBeanOperation operation : operations) {
                    mbean.removeOperation(operation);
                }
            }
        }
    }

    /**
     * Get the BaseModelMBeanExt with the given id.
     * @param id The id of the BaseModelMBeanExt to get.
     * @return The BaseModelMBeanExt with the given id.
     */
    public synchronized IBaseModelMBeanExt getBaseModelMBeanExt(final String id) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        return this.ids2mbeans.get(id);
    }

    /**
     * Register a new BaseModelMBeanExt.
     * @param id The id to register the BaseModelMBeanExt with.
     * @param mbean The BaseModelMBeanExt to register.
     */
    public synchronized void registerBaseModelMBeanExt(final String id, final IBaseModelMBeanExt mbean) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        if (id == null) {
            throw new IllegalArgumentException();
        }

        unregisterBaseModelMBeanExt(id);

        this.ids2mbeans.put(id, mbean);

        Map<IMBeanAttribute, Set<String>> attributes2ids = this.mbeans2attributes2ids.get(mbean);
        if (attributes2ids == null) {
            attributes2ids = new HashMap<IMBeanAttribute, Set<String>>();
            this.mbeans2attributes2ids.put(mbean, attributes2ids);
        }

        Map<IMBeanOperation, Set<String>> operations2ids = this.mbeans2operations2ids.get(mbean);
        if (operations2ids == null) {
            operations2ids = new HashMap<IMBeanOperation, Set<String>>();
            this.mbeans2operations2ids.put(mbean, operations2ids);
        }

        Map<String, Set<IMBeanAttribute>> ids2attributes = this.mbeans2ids2attributes.get(mbean);
        if (ids2attributes == null) {
            ids2attributes = new HashMap<String, Set<IMBeanAttribute>>();
            this.mbeans2ids2attributes.put(mbean, ids2attributes);
        }

        Map<String, Set<IMBeanOperation>> ids2operations = this.mbeans2ids2operations.get(mbean);
        if (ids2operations == null) {
            ids2operations = new HashMap<String, Set<IMBeanOperation>>();
            this.mbeans2ids2operations.put(mbean, ids2operations);
        }

        final Set<IMBeanAttribute> attributes = new HashSet<IMBeanAttribute>();
        ids2attributes.put(id, attributes);

        final Set<IMBeanOperation> operations = new HashSet<IMBeanOperation>();
        ids2operations.put(id, operations);

        for (Entry<IMBeanAttribute, String> entry : this.attributes2regexps.entrySet()) {
            final IMBeanAttribute attribute = entry.getKey();
            final String regexp = entry.getValue();

            if (id.matches(regexp)) {
                Set<String> ids = attributes2ids.get(attribute);

                if (ids == null) {
                    ids = new HashSet<String>();
                    attributes2ids.put(attribute, ids);
                    mbean.addAttribute(attribute);
                }

                ids.add(id);
                attributes.add(attribute);
            }
        }

        for (Entry<IMBeanOperation, String> entry : this.operations2regexps.entrySet()) {
            final IMBeanOperation operation = entry.getKey();
            final String regexp = entry.getValue();

            if (id.matches(regexp)) {
                Set<String> ids = operations2ids.get(operation);

                if (ids == null) {
                    ids = new HashSet<String>();
                    operations2ids.put(operation, ids);
                    mbean.addOperation(operation);
                }

                ids.add(id);
                operations.add(operation);
            }
        }
    }

    /**
     * Unregister a BaseModelMBeanExt.
     * @param id The id to unregister.
     */
    public synchronized void unregisterBaseModelMBeanExt(final String id) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        if (id == null) {
            throw new IllegalArgumentException();
        }

        final IBaseModelMBeanExt mbean = this.ids2mbeans.remove(id);

        if (mbean != null) {
            final Map<String, Set<IMBeanAttribute>> ids2attributes = this.mbeans2ids2attributes.get(mbean);
            final Map<String, Set<IMBeanOperation>> ids2operations = this.mbeans2ids2operations.get(mbean);
            final Map<IMBeanAttribute, Set<String>> attributes2ids = this.mbeans2attributes2ids.get(mbean);
            final Map<IMBeanOperation, Set<String>> operations2ids = this.mbeans2operations2ids.get(mbean);

            final Set<IMBeanAttribute> attributes = ids2attributes.remove(id);
            if (ids2attributes.size() == 0) {
                this.mbeans2ids2attributes.remove(mbean);
                this.mbeans2attributes2ids.remove(mbean);
            }

            final Set<IMBeanOperation> operations = ids2operations.remove(id);
            if (ids2operations.size() == 0) {
                this.mbeans2ids2operations.remove(mbean);
                this.mbeans2operations2ids.remove(mbean);
            }

            for (IMBeanAttribute attribute : attributes) {
                final Set<String> ids = attributes2ids.get(attribute);

                ids.remove(id);
                if (ids.size() == 0) {
                    attributes2ids.remove(attribute);
                    mbean.removeAttribute(attribute);
                }
            }

            for (IMBeanOperation operation : operations) {
                final Set<String> ids = operations2ids.get(operation);

                ids.remove(id);
                if (ids.size() == 0) {
                    operations2ids.remove(operation);
                    mbean.removeOperation(operation);
                }
            }
        }
    }

    /**
     * Register a new attribute.<br>
     * The attribute will automatically be register with each registered BaseModelMBeanExt matching the given regexp.
     * @param attribute The attribute to register.
     * @param regexp The regexp to apply.
     */
    public synchronized void registerAttribute(final IMBeanAttribute attribute, final String regexp) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        unregisterAttribute(attribute);

        this.attributes2regexps.put(attribute, regexp);

        for (Entry<String, IBaseModelMBeanExt> entry : this.ids2mbeans.entrySet()) {
            final String id = entry.getKey();
            final IBaseModelMBeanExt mbean = entry.getValue();

            if (id.matches(regexp)) {
                final Map<String, Set<IMBeanAttribute>> ids2attributes = this.mbeans2ids2attributes.get(mbean);
                final Map<IMBeanAttribute, Set<String>> attributes2ids = this.mbeans2attributes2ids.get(mbean);

                Set<IMBeanAttribute> attributes = ids2attributes.get(id);
                Set<String> ids = attributes2ids.get(attribute);

                if (ids == null) {
                    ids = new HashSet<String>();
                    attributes2ids.put(attribute, ids);
                    mbean.addAttribute(attribute);
                }

                ids.add(id);
                attributes.add(attribute);
            }
        }
    }

    /**
     * Unregister an attribute.
     * @param attribute The attribute to unregister.
     */
    public synchronized void unregisterAttribute(final IMBeanAttribute attribute) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        final String regexp = this.attributes2regexps.remove(attribute);

        if (regexp != null) {
            for (Entry<String, IBaseModelMBeanExt> entry : this.ids2mbeans.entrySet()) {
                final String id = entry.getKey();
                final IBaseModelMBeanExt mbean = entry.getValue();

                if (id.matches(regexp)) {
                    final Map<String, Set<IMBeanAttribute>> ids2attributes = this.mbeans2ids2attributes.get(mbean);
                    final Map<IMBeanAttribute, Set<String>> attributes2ids = this.mbeans2attributes2ids.get(mbean);
                    final Set<IMBeanAttribute> attributes = ids2attributes.get(id);
                    final Set<String> ids = attributes2ids.get(attribute);

                    ids.remove(id);
                    attributes.remove(attribute);

                    if (ids.size() == 0) {
                        attributes2ids.remove(attribute);
                        mbean.removeAttribute(attribute);
                    }
                }
            }
        }
    }

    /**
     * Register a new operation.<br>
     * The operation will automatically be register with each registered BaseModelMBeanExt matching the given regexp.
     * @param operation The operation to register.
     * @param regexp The regexp to apply.
     */
    public synchronized void registerOperation(final IMBeanOperation operation, final String regexp) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        unregisterOperation(operation);

        this.operations2regexps.put(operation, regexp);

        for (Entry<String, IBaseModelMBeanExt> entry : this.ids2mbeans.entrySet()) {
            final String id = entry.getKey();
            final IBaseModelMBeanExt mbean = entry.getValue();

            if (id.matches(regexp)) {
                final Map<String, Set<IMBeanOperation>> ids2operations = this.mbeans2ids2operations.get(mbean);
                final Map<IMBeanOperation, Set<String>> operations2ids = this.mbeans2operations2ids.get(mbean);

                Set<IMBeanOperation> operations = ids2operations.get(id);
                Set<String> ids = operations2ids.get(operation);

                if (ids == null) {
                    ids = new HashSet<String>();
                    operations2ids.put(operation, ids);
                    mbean.addOperation(operation);
                }

                ids.add(id);
                operations.add(operation);
            }
        }
    }

    /**
     * Unregister an operation.
     * @param operation The operation to unregister.
     */
    public synchronized void unregisterOperation(final IMBeanOperation operation) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        final String regexp = this.operations2regexps.remove(operation);

        if (regexp != null) {
            for (Entry<String, IBaseModelMBeanExt> entry : this.ids2mbeans.entrySet()) {
                final String id = entry.getKey();
                final IBaseModelMBeanExt mbean = entry.getValue();

                if (id.matches(regexp)) {
                    final Map<String, Set<IMBeanOperation>> ids2operations = this.mbeans2ids2operations.get(mbean);
                    final Map<IMBeanOperation, Set<String>> operations2ids = this.mbeans2operations2ids.get(mbean);
                    final Set<IMBeanOperation> operations = ids2operations.get(id);
                    final Set<String> ids = operations2ids.get(operation);

                    ids.remove(id);
                    operations.remove(operation);

                    if (ids.size() == 0) {
                        operations2ids.remove(operation);
                        mbean.removeOperation(operation);
                    }
                }
            }
        }
    }
}
